package ru.dbotthepony.kstarbound.server.world

import it.unimi.dsi.fastutil.longs.Long2ObjectFunction
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectAVLTreeSet
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import ru.dbotthepony.kommons.core.KOptional
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.server.StarboundServer
import ru.dbotthepony.kstarbound.server.network.ServerConnection
import ru.dbotthepony.kstarbound.util.composeFutures
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.WorldGeometry
import java.io.Closeable
import java.util.concurrent.CompletableFuture
import java.util.concurrent.atomic.AtomicInteger
import java.util.concurrent.locks.LockSupport
import java.util.function.Consumer
import kotlin.concurrent.withLock

class ServerWorld(
	val server: StarboundServer,
	seed: Long,
	geometry: WorldGeometry,
) : World<ServerWorld, ServerChunk>(seed, geometry) {
	init {
		server.worlds.add(this)
	}

	val players = ObjectArraySet<ServerConnection>()

	val thread = Thread(::runThread, "Starbound Server World $seed")
	var isStopped: Boolean = false
		private set

	init {
		thread.isDaemon = true
	}

	private val chunkProviders = ArrayList<IChunkSource>()
	var saver: IChunkSaver? = null

	fun addChunkSource(source: IChunkSource) {
		chunkProviders.add(source)
	}

	@Volatile
	private var nextThink = 0L

	override fun close() {
		super.close()
		isStopped = true
	}

	fun startThread() {
		nextThink = System.nanoTime()
		thread.start()
	}

	private fun runThread() {
		while (!isStopped) {
			var diff = System.nanoTime() - nextThink

			while (diff < Starbound.TICK_TIME_ADVANCE_NANOS) {
				mailbox.executeQueuedTasks()
				diff = System.nanoTime() - nextThink
				LockSupport.parkNanos(diff)
				diff = System.nanoTime() - nextThink
			}

			nextThink = System.nanoTime()

			try {
				think()
			} catch (err: Throwable) {
				close()
				throw err
			}
		}
	}

	override val isRemote: Boolean
		get() = false

	override fun thinkInner() {
		lock.withLock {
			players.forEach { it.tick() }

			ticketLists.removeIf {
				val valid = it.tick()

				if (!valid) {
					val removed = ticketMap.remove(it.pos.toLong())
					check(removed == it) { "Expected to remove $it, but removed $removed" }

					val chunk = chunkMap[it.pos]

					if (chunk != null) {
						saver?.saveCells(it.pos, chunk.copyCells())
						chunkMap.remove(it.pos)
					}
				}

				!valid
			}
		}
	}

	override fun chunkFactory(pos: ChunkPos): ServerChunk {
		return ServerChunk(this, pos)
	}

	private val ticketMap = Long2ObjectOpenHashMap<TicketList>()
	private val ticketLists = ArrayList<TicketList>()

	private fun getTicketList(pos: ChunkPos): TicketList {
		return ticketMap.computeIfAbsent(geometry.wrapToLong(pos), Long2ObjectFunction { TicketList(it) })
	}

	fun permanentChunkTicket(pos: ChunkPos): ITicket {
		lock.withLock {
			return getTicketList(pos).Ticket()
		}
	}

	fun temporaryChunkTicket(pos: ChunkPos, time: Int): ITicket {
		require(time > 0) { "Invalid ticket time: $time" }

		lock.withLock {
			return getTicketList(pos).TimedTicket(time)
		}
	}

	interface ITicket {
		fun cancel()
		val isCanceled: Boolean
		val pos: ChunkPos
		val id: Int
	}

	interface ITimedTicket : ITicket, Comparable<ITimedTicket> {
		val timeRemaining: Int
		fun prolong(ticks: Int)

		override fun compareTo(other: ITimedTicket): Int {
			val cmp = timeRemaining.compareTo(other.timeRemaining)
			if (cmp != 0) return cmp
			return id.compareTo(other.id)
		}
	}

	private inner class TicketList(val pos: ChunkPos) {
		constructor(pos: Long) : this(ChunkPos(pos))

		private var first = true
		private val permanent = ArrayList<Ticket>()
		private val temporary = ObjectAVLTreeSet<TimedTicket>()
		private var ticks = 0
		private var nextTicketID = AtomicInteger()

		val isValid: Boolean
			get() = temporary.isNotEmpty() || permanent.isNotEmpty()

		fun tick(): Boolean {
			ticks++

			while (temporary.isNotEmpty() && temporary.first().timeRemaining <= 0) {
				val ticket = temporary.first()
				ticket.isCanceled = true
				temporary.remove(ticket)
			}

			return temporary.isNotEmpty() || permanent.isNotEmpty()
		}

		abstract inner class AbstractTicket : ITicket {
			final override val id: Int = nextTicketID.getAndIncrement()
			final override val pos: ChunkPos
				get() = this@TicketList.pos

			final override var isCanceled: Boolean = false

			fun init() {
				if (first) {
					first = false

					if (geometry.x.inBoundsChunk(pos.x) && geometry.y.inBoundsChunk(pos.y)) {
						ticketLists.add(this@TicketList)

						if (chunkProviders.isNotEmpty()) {
							composeFutures(chunkProviders)
							{ if (!isValid) CompletableFuture.completedFuture(KOptional.empty()) else it.getTiles(pos) }
								.thenAccept(Consumer { tiles ->
									if (!isValid || !tiles.isPresent) return@Consumer

									composeFutures(chunkProviders)
									{ if (!isValid) CompletableFuture.completedFuture(KOptional.empty()) else it.getEntities(pos) }
										.thenAcceptAsync(Consumer { ents ->
											if (!isValid) return@Consumer
											val chunk = chunkMap.compute(pos) ?: return@Consumer
											chunk.loadCells(tiles.value)

											ents.ifPresent {
												for (obj in it) {
													chunk.addObject(obj)
												}
											}
										}, mailbox)
								})
						}
					}
				}
			}

			final override fun cancel() {
				if (isCanceled) return

				lock.withLock {
					if (isCanceled) return
					isCanceled = true
					onCancel()
				}
			}

			protected abstract fun onCancel()
		}

		inner class Ticket : AbstractTicket() {
			init {
				permanent.add(this)
				init()
			}

			override fun onCancel() {
				permanent.remove(this)
			}
		}

		inner class TimedTicket(expiresAt: Int) : AbstractTicket(), ITimedTicket {
			var expiresAt = expiresAt + ticks

			override val timeRemaining: Int
				get() = (expiresAt - ticks).coerceAtLeast(0)

			init {
				temporary.add(this)
				init()
			}

			override fun onCancel() {
				temporary.remove(this)
			}

			override fun prolong(ticks: Int) {
				if (ticks == 0 || isCanceled) return

				lock.withLock {
					if (isCanceled) return

					temporary.remove(this)
					expiresAt += ticks
					if (timeRemaining > 0) temporary.add(this)
				}
			}
		}
	}
}
